from argparse import ArgumentParser

def main():

  parser = ArgumentParser()
  parser.add_argument('-n', '--num-samples', type=int, default=int(1e6))
  parser.add_argument('-l', '--max-len', type=int, default=150)
  parser.add_argument('-s', '--seed', type=int, default=None)
  parser.add_argument('-v', '--verbose', action='store_true')
  parser.add_argument('-r', '--replace', action='store_true')
  parser.add_argument('-p', '--num-procs', type=int, default=None)

  io_group = parser.add_mutually_exclusive_group(required=True)
  files_spec = io_group.add_argument_group()
  files_spec.add_argument('-i', '--inputs', type=str)
  files_spec.add_argument('-u', '--ucis', type=str)

  io_group.add_argument('-o', '--output', type=str)

  args = parser.parse_args()

  import os

  if args.output:
    os.makedirs(args.output, exist_ok=True)
    input_path = os.path.join(args.output, 'inputs.txt')
    uci_path = os.path.join(args.output, 'ucis.txt')
  else:
    input_path = args.inputs
    uci_path = args.ucis

  import sys
  from pgn_clis.lib.random_samples import random_samples

  mode = 'w' if args.replace else 'x'
  with open(input_path, mode) as inputs_f, open(uci_path, mode) as ucis_f:
    random_samples(
      inputs_f, ucis_f,
      num_samples=args.num_samples, max_len=args.max_len, num_procs=args.num_procs,
      seed=args.seed, logstream=sys.stderr if args.verbose else None
    )


from typing import Sequence, NamedTuple, Iterable, TextIO
import os
import multiprocessing as mp
import random
import chess_utils as cu
import chess

class Sample(NamedTuple):
  inputs: Sequence[str]
  ucis: Sequence[str]

def random_sample(max_len: int = 150, rng = random.Random()) -> Sample:
  board = chess.Board()
  inputs = []
  ucis = []

  for _ in range(max_len):
    move = cu.random.move(board, rng)
    if not move:
      break
    san = board.san(move)
    board.push(move)
    inputs.append(san)
    ucis.append(move.uci())

  return Sample(inputs, ucis)


def random_samples(
  inputs: TextIO, ucis: TextIO, *,
  num_samples: int, max_len: int = 150,
  seed: int | None = None, logstream: TextIO | None = None,
  num_procs: int | None = None
):
  num_procs = num_procs or os.cpu_count() or 1
  proc_samples = num_samples // num_procs

  if logstream:
    print(f'Generating {num_samples} samples...', file=logstream)
    print(f'  max_len: {max_len}', file=logstream)
    print(f'  num_procs: {num_procs}', file=logstream)
    print(f'  seed: {seed or "random"}', file=logstream)
    print()

  q = mp.Queue()

  def collector(
    inputs: TextIO, ucis: TextIO,
    *, queue: mp.Queue,
    logstream: TextIO | None = None
  ):
    i = 0
    while True:
      data = queue.get()
      if data is None:
        break
      inputs.write(' '.join(data.inputs) + '\n')
      ucis.write(' '.join(data.ucis) + '\n')

      if logstream:
        i += 1
        if i % 100 == 0:
          print(f'\r{i} / {num_samples} ({100*i/num_samples:.2f}%)', end='', file=logstream)

  def generator(queue: mp.Queue, num_samples: int, seed: int | None):
    rng = random.Random(seed)
    for _ in range(num_samples):
      sample = random_sample(max_len, rng)
      queue.put(sample)

  collector_proc = mp.Process(target=collector, args=(inputs, ucis), kwargs=dict(queue=q, logstream=logstream))
  collector_proc.start()

  gen_procs: list[mp.Process] = []
  for _ in range(num_procs):
    gen_proc = mp.Process(target=generator, args=(q, proc_samples, seed))
    gen_proc.start()
    gen_procs.append(gen_proc)

  for gen_proc in gen_procs:
    gen_proc.join()

  q.put(None)
  collector_proc.join()
